// Copyright © 2021 Charles Kerr. All rights reserved.
// Created on:  5/19/21

#include "DataCategory.hpp"
#include "StringUtility.hpp"
#include <cstdint>
#include <iostream>
#include <stdexcept>
#include <fstream>

//+++++++++++++++++++++++++++++++++++++++++++++++++++
//
//		Methods for DataTypeStorage
//
//+++++++++++++++++++++++++++++++++++++++++++++++++++
const  std::map<DEFINITIONCATEGORIES,std::string> DataCategory::directories = {
	{items_def,"items"},
	{npc_def,"npc"},
	{create_def,"create"},
	{regions_def,"regions"},
	{misc_def,"misc"},
	{skills_def,"skills"},
	{location_def,"location"},
	{menus_def,"menus"},
	{spells_def,"spells"},
	{newbie_def,"newbie"},
	{titles_def,"titles"},
	{advance_def,"advance"},
	{house_def,"house"},
	{colors_def,"colors"},
	{spawn_def,"spawn"},
	{html_def,"html"},
	{race_def,"race"},
	{weathab_def,"weather"},
	{hard_items_def,"harditems"},
	{command_def,"command"},
	{msgboard_def,"msgboard"},
	{carve_def,"carve"},
	{creatures_def,"creatures"},
	{maps_def,"maps"}
};



//=============================================================================
std::string DataCategory::retrieveSectionName(const std::string& data){
	auto start = data.find("[");
	auto end = data.find("]");
	if ((start== std::string::npos) || (end == std::string::npos)){
		throw std::runtime_error(std::string("Invalid section header: ")+data);
	}
	return strutil::trim(data.substr(start+1,end-start-1));
}

//=============================================================================
std::shared_ptr<KeyValue> DataCategory::processLine(const std::string line,const std::string &filepath, int linenumber){
	std::shared_ptr<KeyValue> rvalue = std::make_shared<KeyValue>();
	rvalue->original_line = line ;
	rvalue->fileorigin=filepath ;
	rvalue->linenumber = linenumber ;
	auto parsed = parseKeyValue(line);
	auto key = std::get<0>(parsed) ;
	auto ukey = strutil::toupper(key);
	auto value = std::get<1>(parsed) ;
	rvalue->ukey = ukey;
	rvalue->key = key ;
	rvalue->value = value ;
	auto etag = KeyValue::lookupKeyEnum(ukey);
	rvalue->etag = etag ;
	rvalue->tag = key ;
	rvalue->textdata = value ;
	
	auto type = KeyValue::lookupKeyType(etag);
	auto numbers = convertToInt(value);
	switch (_mycategory) {
		case advance_def:
		case hard_items_def:
		case npc_def:
		case items_def: {
			if (value.empty() && (type !=KeyValueType::nodata)){
				return nullptr ;
			}

			switch (etag) {
					
				case DFNTAG_ITEMLIST:
				case DFNTAG_NPCLIST:
				case DFNTAG_INVALIDTAG:
					break;
				default:
				{
					switch (type) {
						case KeyValueType::unknown:
						case KeyValueType::nodata:
							rvalue->textdata = "";
							break;
							
						case KeyValueType::text:
							rvalue->textdata = value ;
							break;
							
						case KeyValueType::uppertext:
							rvalue->textdata = strutil::toupper(value);
							break;
							
						case KeyValueType::numeric:
							rvalue->firstnumber = std::get<0>(numbers);
							break;
							
						case KeyValueType::doublenumeric:
							rvalue->firstnumber = std::get<0>(numbers);
							rvalue->firstnumber = std::get<1>(numbers);
							break;
							
						default:
							rvalue->textdata = value;
							break;
					}
					
					break;
				}
			}
			break;
		}
		case spawn_def:
		case create_def:
		case command_def:
			key = ukey;
			rvalue->tag = key ;
		default:
			rvalue->textdata = value ;
			break;
	}
	
	return rvalue ;
}
//=============================================================================
std::map<std::string,int> DataCategory::retreivePostFiles(){
	auto filepath = _basedir / std::filesystem::path("priority.nfo") ;
	std::map<std::string,int> rvalue ;
	if (!std::filesystem::exists(filepath)) {
		return rvalue ;
	}
	// We have a priority file
	auto input = std::ifstream(filepath.string(),std::ios::in) ;
	if (!input.is_open()){
		return rvalue ;
	}
	char buffer[2049];
	int priority = 0 ;
	while (!input.eof() && !input.fail()){
		input.getline(buffer,2048);
		if (input.gcount() > 0) {
			buffer[input.gcount()] = 0 ;
			auto data = strutil::trim(buffer);
			if (!data.empty()) {
				auto filename = strutil::toupper(data) ;
				rvalue.insert_or_assign(filename, priority);
			}
			
		}
	}
	input.close();
	return rvalue ;
}

//=============================================================================
std::tuple<std::string,std::string> DataCategory::parseKeyValue(const std::string& data,const std::string &sep) {
	std::string key ;
	std::string value ;
	key = data ;
	auto loc = data.find(sep) ;
	if (loc != std::string::npos) {
		key = strutil::trim(data.substr(0,loc));
		if (loc+1 < data.size()) {
			value = strutil::trim(data.substr(loc+1)) ;
		}
		
	}
	return std::make_tuple(key,value);
}

//=============================================================================
std::tuple<int,int> DataCategory::convertToInt(const std::string& value, const std::string &sep){
	int first = 0 ;
	int second = 0 ;
	auto loc= value.find(sep) ;
	if (loc != std::string::npos){
		try {
			first = std::stoi(value.substr(0,loc),nullptr,0) ;
		}
		catch(...){
			first = 0 ;
		}
		if ((loc+1)<value.size() ){
			try {
				second = std::stoi(value.substr(loc+1),nullptr,0);
			}
			catch(...){
				second = first;
			}
		}
		else {
			second = first ;
		}
	}
	return std::make_tuple(first,second);
}

//=============================================================================
void  DataCategory::loadFile(const std::string& filepath){
	
	int linenumber = 0;
	std::string section_name ;
	std::shared_ptr<DataSection> data_section = nullptr ;
	bool sectionheader = false ;
	bool insection = false ;
	auto input = std::ifstream(filepath);
	if (!input.is_open()){
		return  ;
	}
	char buffer[2049];
	while (!input.eof() && !input.fail()){
		input.getline(buffer,2048);
		if (input.gcount()>0) {
			buffer[input.gcount()]=0;
			auto line = std::string(buffer) ;
			strutil::stripTrim(line);
			if (!line.empty()) {
				if (line[0] == '[')  {
					if (insection || sectionheader) {
						throw std::runtime_error(std::string("Section header found while in section or after sectionheader '")+section_name+std::string("' at line ")+std::to_string(linenumber)+std::string(" in file: ")+filepath);
					}
					try {
						section_name = retrieveSectionName(line);
					}
					catch(const std::runtime_error &e) {
						throw std::runtime_error(std::string("Invalid header found at line: ")+std::to_string(linenumber)+ std::string(" in file: ")+filepath+std::string(" header was: ")+line);
					}
					sectionheader = true ;
					
				}
				else if (line[0] == '{') {
					if (!sectionheader) {
						throw std::runtime_error(std::string("Section start found before section header at line: ") +std::to_string(linenumber) + std::string(" in file: ")+filepath);
					}
					insection = true ;
					data_section = std::make_shared<DataSection>(section_name,filepath,linenumber);
				}
				else if (line[0] == '}') {
					if (!sectionheader || (data_section == nullptr)){
						throw std::runtime_error(std::string("Section end found before section header at line: ") +std::to_string(linenumber) + std::string(" in file: ")+filepath);

					}
					_sections.insert_or_assign(strutil::toupper(data_section->name()), data_section);
					data_section = nullptr;
					sectionheader = false ;
					insection = false ;
					
				}
				else {
					if (insection && sectionheader && (data_section != nullptr)) {
						// process the data
						auto value = processLine(line);
						if (value != nullptr) {
							value->fileorigin=filepath ;
							value->linenumber = linenumber ;
							data_section->addKeyValue(value);
						}
					}
				}
			}
			
		}
		linenumber++;
	}
	if (insection){
		throw std::runtime_error(std::string("End of file before closing section found: ")+filepath);
	}
	
}

//=============================================================================
std::shared_ptr<DataSection> DataCategory::section(const std::string& name){
	auto iter = _sections.find(strutil::toupper(name));
	if (iter != _sections.end()) {
		return iter->second;
	}
	return nullptr ;
}

//=============================================================================
std::shared_ptr<DataSection> DataCategory::containsName(const std::string& name){
	for (auto &iter : _sections) {
		if (iter.first.find(strutil::toupper(name))!= std::string::npos) {
			return iter.second;
		}
	}
	return nullptr;
}


//=============================================================================
void DataCategory::loadData(const std::string &directory, DEFINITIONCATEGORIES category,const std::string &extension){
	_sections.clear();
	auto path = std::filesystem::path(directory) ;
	_basedir = path ;
	auto cat = directories.find(category);
	if (cat == directories.end()) {
		throw std::runtime_error(std::string("No subdirectory defined for category number: ")+std::to_string(category));
	}
	auto subdir = cat->second ;
	if (!subdir.empty()) {
		_basedir = path / std::filesystem::path(subdir) ;
	}
	if (!std::filesystem::exists(_basedir)) {
		throw std::runtime_error(std::string("Data category directory does not exist: ")+_basedir.string());
	}
	auto priorities = retreivePostFiles() ;
	// recursive all subdirectories and files
	auto postfiles = std::vector<std::filesystem::path>() ;
	for (auto &file : std::filesystem::recursive_directory_iterator(_basedir)) {
		auto tempname = file.path().string();
		if (std::filesystem::is_regular_file(file.path())) {
			if (file.path().extension() == extension) {
				// One we may be intersted in
				auto filename = strutil::toupper( file.path().filename().string() );
				auto iter = priorities.find(filename) ;
				if (iter!= priorities.end()) {
					postfiles.push_back(file.path());
				}
				else {
					// Process the file
					loadFile(file.path().string());
					
				}
			}
		}
	}
	for (auto &path : postfiles) {
		loadFile(path.string());
	}
	
}

//=============================================================================
std::size_t DataCategory::size() const  {
	return _sections.size();
}
//=============================================================================
std::map<std::string,std::shared_ptr<DataSection> > DataCategory::sections() const {
	return _sections;
}

//=============================================================================
DEFINITIONCATEGORIES DataCategory::category() const {
	return _mycategory;
}
//=============================================================================
void DataCategory::setCategory(DEFINITIONCATEGORIES value) {
	_mycategory = value ;
}

//=============================================================================
DataCategory::DataCategory(){
	_mycategory = invalid_cat ;
}

//=============================================================================
DataCategory::DataCategory(const std::string &filepath): DataCategory(){
	loadFile(filepath);
}
//============================================================================
void DataCategory::clear() {
	_sections.clear();
}
